@@ -12,18 +12,19 @@ from pathlib import Path
 from typing import Any, Callable
 
 import tqdm
+from optimade import __api_version__ as OPTIMADE_API_VERSION
 from optimade.models import EntryInfoResource, EntryResource
 from optimade.server.schemas import ENTRY_INFO_SCHEMAS, retrieve_queryable_properties
 
+from .aiida_plugin import AiidaEntryPath, construct_entries_from_aiida
 from .config import Config, EntryConfig, JSONLConfig, ParsedFiles, PropertyDefinition
-from .parsers import ENTRY_PARSERS, OPTIMADE_CONVERTERS, PROPERTY_PARSERS, TYPE_MAP
 
-PROVIDER_PREFIX = os.environ.get("optimake_PROVIDER_PREFIX", "optimake")
+PROVIDER_PREFIX = os.environ.get("OPTIMAKE_PROVIDER_PREFIX", "optimake")
 
 
 def _construct_entry_type_info(
     type: str,
-    properties: list[PropertyDefinition],
+    properties: list[PropertyDefinition] | list[dict],
     provider_prefix: str,
 ) -> EntryInfoResource:
     """Take the provided property definitions and construct an entry info response.
@@ -40,28 +41,42 @@ def _construct_entry_type_info(
         )
 
     info: dict[str, Any] = {"formats": ["json"], "description": type}
-    info["properties"] = {
-        f"_{provider_prefix}_{p.name}": {
-            "description": p.description,
-            "unit": p.unit,
-            "type": p.type,
-            "title": p.title,
+    info["properties"] = {}
+    for p in properties:
+        if isinstance(p, PropertyDefinition):
+            p = p.model_dump()
+
+        p_name = (
+            f"_{provider_prefix}_{p['name']}"
+            if not p["name"].startswith(f"_{provider_prefix}")
+            else p["name"]
+        )
+        info["properties"][p_name] = {
+            "description": p.get("description"),
+            "unit": p.get("unit"),
+            "type": p.get("type"),
+            "title": p.get("title"),
         }
-        for p in properties
-    }
+
     info["properties"].update(default_properties)
     info["output_fields_by_format"] = {}
     info["output_fields_by_format"]["json"] = list(info["properties"].keys())
     return EntryInfoResource(**info)
 
 
-def convert_archive(archive_path: Path, jsonl_path: Path | None = None) -> Path:
-    """Convert an MCloud entry to an OPTIMADE JSONL file.
+def convert_archive(
+    archive_path: Path,
+    jsonl_path: Path | None = None,
+    limit: int | None = None,
+    overwrite: bool = False,
+) -> Path:
+    """Convert a raw data archive to an OPTIMADE JSONL file.
 
     Parameters:
         archive_path: The location of the `optimade.yaml` file to convert.
         jsonl_path: The location to write the JSONL file to. If not provided,
             write to `<archive_path>/optimade.jsonl`.
+        limit: The maximum number of entries to parse (useful for testing).
 
     Raises:
         FileNotFoundError: If any of the data paths in the config file,
@@ -75,6 +90,8 @@ def convert_archive(archive_path: Path, jsonl_path: Path | None = None) -> Path:
 
     if not jsonl_path:
         jsonl_path = archive_path / "optimade.jsonl"
+    if jsonl_path.exists() and not overwrite:
+        raise RuntimeError(f"Not overwriting existing file at {jsonl_path}")
 
     # if the config specifies just a JSON-L, then extract any archives
     # and return the JSONL path
@@ -85,28 +102,33 @@ def convert_archive(archive_path: Path, jsonl_path: Path | None = None) -> Path:
         if jsonl_path != src_jsonl_path:
             # add a symlink to the specified jsonl_path
             if jsonl_path.exists():
-                raise RuntimeError(f"Not overwriting existing file at {jsonl_path}")
+                if overwrite:
+                    jsonl_path.unlink()
+                else:
+                    raise RuntimeError(f"Not overwriting existing file at {jsonl_path}")
             jsonl_path.symlink_to(archive_path / src_jsonl_path)
         return jsonl_path
 
-    # first, decompress any provided data paths
-    data_paths: set[Path] = set()
+    # first, decompress any provided data archives
+    decompress_paths: set[Path] = set()
     for entry in mc_config.entries:
-        for e in entry.entry_paths:
-            if e.matches:
-                data_paths.add((archive_path / str(e.file)).resolve())
-        for p in entry.property_paths:
-            if p.matches:
-                data_paths.add((archive_path / str(p.file)).resolve())
+        for entry_path in entry.entry_paths:
+            if getattr(entry_path, "matches", None):
+                decompress_paths.add((archive_path / str(entry_path.file)).resolve())
+        for prop_path in entry.property_paths:
+            if getattr(prop_path, "matches", None):
+                decompress_paths.add((archive_path / str(prop_path.file)).resolve())
 
-    for data_path in data_paths:
-        inflate_archive(archive_path, data_path)
+    for decompress_path in decompress_paths:
+        inflate_archive(archive_path, decompress_path)
 
     optimade_entries: dict[str, list[dict]] = defaultdict(list)
 
     for entry in mc_config.entries:
         optimade_entries[entry.entry_type].extend(
-            construct_entries(archive_path, entry, PROVIDER_PREFIX).values()
+            construct_entries(
+                archive_path, entry, PROVIDER_PREFIX, limit=limit
+            ).values()
         )
 
     property_definitions = defaultdict(list)
@@ -119,6 +141,7 @@ def convert_archive(archive_path: Path, jsonl_path: Path | None = None) -> Path:
         property_definitions,
         PROVIDER_PREFIX,
         jsonl_path,
+        overwrite,
     )
 
     return jsonl_path
@@ -227,21 +250,35 @@ def _parse_entries(
     archive_path: Path,
     matches_by_file: dict[str | None, list[Path]],
     entry_type: str,
+    limit: int | None = None,
 ) -> tuple[list[Any], list[str]]:
     """Loop through the matches by file and parse them into
     the intermediate format, also generating IDs for each.
 
+    Parameters:
+        archive_path: The path to the archive.
+        matches_by_file: A dictionary of matches by file.
+        entry_type: The type of entry to parse.
+        limit: The maximum number of entries to parse
+
     Returns:
         A list of parsed entries and a list of IDs.
 
     """
+    from .parsers import ENTRY_PARSERS
+
     parsed_entries = []
     entry_ids: list[str] = []
     for archive_file in matches_by_file:
-        for _path in tqdm.tqdm(
-            matches_by_file[archive_file],
-            desc=f"Parsing {entry_type} files",
+        for ind, _path in enumerate(
+            tqdm.tqdm(
+                matches_by_file[archive_file],
+                desc=f"Parsing {entry_type} files",
+            )
         ):
+            if limit and ind >= limit:
+                break
+
             path_in_archive: Path = Path(_path).relative_to(Path(archive_path))
             exceptions = {}
 
@@ -294,6 +331,9 @@ def _set_unique_entry_ids(entry_ids: list[str]) -> list[str]:
 
     """
 
+    if len(entry_ids) == 1:
+        return [os.path.splitext(os.path.basename(entry_ids[0]))[0]]
+
     new_ids: list[str] = list(entry_ids)
 
     def _strip_common_path(ids, from_back=False):
@@ -340,28 +380,23 @@ def _set_unique_entry_ids(entry_ids: list[str]) -> list[str]:
     return new_ids
 
 
-def _parse_and_assign_properties(
-    optimade_entries: dict[str, EntryResource],
+def _parse_properties_from_files(
     property_matches_by_file: dict[str | None, list[Path]],
-    entry_type: str,
     property_definitions: list[PropertyDefinition],
-    provider_prefix: str,
-) -> None:
-    """Loop through the property matches by file and parse them into the combined
-    dictionary of OPTIMADE entries.
+) -> dict[str, dict[str, Any]]:
+    """Loop through the property matches by file and parse them into a dictionary"""
+    from .parsers import PROPERTY_PARSERS
 
-    """
     parsed_properties: dict[str, dict[str, Any]] = defaultdict(dict)
     errors = []
-    all_property_fields: set[str] = set()
 
     if not property_matches_by_file:
-        return
+        return {}
 
     for archive_file in property_matches_by_file:
         for _path in tqdm.tqdm(
             property_matches_by_file[archive_file],
-            desc=f"Parsing properties for {entry_type} entries",
+            desc="Parsing property files",
         ):
             file_ext = _path.suffix
             for parser in PROPERTY_PARSERS[file_ext]:
@@ -369,7 +404,6 @@ def _parse_and_assign_properties(
                     properties = parser(_path, property_definitions)
                     for id in properties:
                         parsed_properties[id].update(properties[id])
-                        all_property_fields |= set(properties[id].keys())
                     break
                 except Exception as exc:
                     errors.append(exc)
@@ -378,50 +412,115 @@ def _parse_and_assign_properties(
                 raise RuntimeError(
                     f"Could not parse properties file {_path} with any of the provided parsers {PROPERTY_PARSERS[file_ext]}. Errors: {errors}"
                 )
-
     if not parsed_properties:
         raise RuntimeError(
             f"Could not parse properties files with any of the provided parsers. Errors: {errors}"
         )
+    return parsed_properties
 
-    # Match properties up to the descrptions provided in the config
+
+def _assign_and_validate_properties(
+    optimade_entries: dict[str, EntryResource],
+    parsed_properties: dict[str, dict[str, Any]],
+    property_definitions: list[PropertyDefinition],
+    provider_prefix: str,
+) -> None:
+    """Assign parsed properties to the OPTIMADE entries."""
+    from .parsers import TYPE_MAP
+
+    # Collect all property fields that are already present in the OPTIMADE entries
+    existing_property_fields: set[str] = set()
+    # Only check the first entry for property fields (it should have all the fields)
+    first_entry = next(iter(optimade_entries.values()), None)
+    if first_entry:
+        prefix = f"_{provider_prefix}_"
+        existing_property_fields |= set(
+            k[len(prefix) :]
+            for k in first_entry["attributes"].keys()
+            if k.startswith(prefix)
+        )
+
+    all_parsed_fields: set[str] = set()
+    for properties in parsed_properties.values():
+        all_parsed_fields |= set(properties.keys())
+
+    # Warn if any parsed property IDs does not exist in the OPTIMADE entries
+    optimade_immutable_ids = {
+        entry["attributes"].get("immutable_id") for entry in optimade_entries.values()
+    }
+    for id in parsed_properties:
+        if id not in optimade_entries and id not in optimade_immutable_ids:
+            warnings.warn(
+                f"Could not find entry {id!r} in OPTIMADE entries.",
+            )
+
+    # Match properties up to the descriptions provided in the config
     property_def_dict: dict[str, PropertyDefinition] = {
         p.name: p for p in property_definitions
     }
-    expected_property_fields = set(property_def_dict.keys())
+
+    if parsed_properties:
+        # Assign parsed properties to the OPTIMADE entries
+        # Look for precisely matching IDs, or 'filename' matches
+        for id in optimade_entries:
+            # detect any other compatible IDs; either those matching immutable ID or those matching the filename rule
+            property_entry_id = optimade_entries[id]["attributes"].get(
+                "immutable_id", None
+            )
+            if property_entry_id is None:
+                # try to find a matching ID based on the filename
+                property_entry_id = id.split("/")[-1].split(".")[0]
+
+            if (property_entry_id not in parsed_properties) and (
+                id not in parsed_properties
+            ):
+                warnings.warn(
+                    f"Could not find entry {id!r} (or fully-qualified {property_entry_id!r}) in parsed properties",
+                )
+                continue
+
+            # Loop over all defined properties and assign them to the entry, setting to None if missing
+            # Also cast types if provided
+            for property in all_parsed_fields:
+                # Look up both IDs: the file path-based ID or the ergonomic one
+                # Different property sources can use different ID schemes internally
+                value = parsed_properties.get(property_entry_id, {}).get(
+                    property, None
+                ) or parsed_properties.get(id, {}).get(property, None)
+                if property not in property_def_dict:
+                    # Fields that are not configured but are present in the property file, will be warned later
+                    continue
+                if value is not None and property_def_dict[property].type in TYPE_MAP:
+                    try:
+                        value = TYPE_MAP[property_def_dict[property].type](value)
+                    except Exception as exc:
+                        raise RuntimeError(
+                            f"Could not cast {value=} for {property=} to type {property_def_dict[property].type!r} for entry {id!r}"
+                        ) from exc
+
+                optimade_entries[id]["attributes"][f"_{provider_prefix}_{property}"] = (
+                    value
+                )
+
+    # Finally, check that all expected property fields are present
+    expected_property_fields = set(p.name for p in property_definitions)
+    all_property_fields = existing_property_fields | all_parsed_fields
 
     if expected_property_fields != all_property_fields:
-        warnings.warn(
-            f"Found {all_property_fields=} in data but {expected_property_fields} in config"
-        )
+        warning_message = "Mismatch between parsed property fields (A) and those defined in config (B)."
+        if all_property_fields - expected_property_fields:
+            warning_message += f"\n(A - B) = {all_property_fields - expected_property_fields} (will be omitted from API; if intended this can be ignored)."
+        if expected_property_fields - all_property_fields:
+            warning_message += f"\n(B - A) = {expected_property_fields - all_property_fields} (configured, but missing; check for typos or missing aliases)"
 
-    # Look for precisely matching IDs, or 'filename' matches
-    for id in optimade_entries:
-        # detect any other compatible IDs; either those matching immutable ID or those matching the filename rule
-        property_entry_id = optimade_entries[id]["attributes"].get("immutable_id", None)
-        if property_entry_id is None:
-            # try to find a matching ID based on the filename
-            property_entry_id = id.split("/")[-1].split(".")[0]
-
-        # Loop over all defined properties and assign them to the entry, setting to None if missing
-        # Also cast types if provided
-        for property in all_property_fields:
-            # Look up both IDs: the file path-based ID or the ergonomic one
-            # Different property sources can use different ID schemes internally
-            value = parsed_properties.get(property_entry_id, {}).get(
-                property, None
-            ) or parsed_properties.get(id, {}).get(property, None)
-            if property not in property_def_dict:
-                warnings.warn(f"Missing property definition for {property=}")
-                continue
-            if value is not None and property_def_dict[property].type in TYPE_MAP:
-                value = TYPE_MAP[property_def_dict[property].type](value)
-
-            optimade_entries[id]["attributes"][f"_{provider_prefix}_{property}"] = value
+        warnings.warn(warning_message)
 
 
-def construct_entries(
-    archive_path: Path, entry_config: EntryConfig, provider_prefix: str
+def construct_entries_from_files(
+    archive_path: Path,
+    entry_config: EntryConfig,
+    provider_prefix: str,
+    limit: int | None = None,
 ) -> dict[str, dict]:
     """Given an archive path and an entry specification,
     loop through the provided paths and try to ingest them
@@ -433,8 +532,8 @@ def construct_entries(
         RuntimeError: If the entry type is not supported.
         ValueError: If any of the files cannot be parsed into
             the given entry type.
-
     """
+    from .parsers import ENTRY_PARSERS, OPTIMADE_CONVERTERS
 
     if entry_config.entry_type not in ENTRY_PARSERS:
         raise RuntimeError(f"Parsing type {entry_config.entry_type} is not supported.")
@@ -453,17 +552,12 @@ def construct_entries(
         archive_path,
         entry_matches_by_file,
         entry_config.entry_type,
+        limit=limit,
     )
 
     # Generate a better set of entry IDs
     unique_entry_ids = _set_unique_entry_ids(file_path_entry_ids)
 
-    # Parse properties
-    property_matches_by_file: dict[str | None, list[Path]] = _get_matches(
-        archive_path, entry_config.property_paths
-    )
-    _check_missing(property_matches_by_file)
-
     timestamp = datetime.datetime.now().isoformat()
 
     # Construct OPTIMADE entries from intermediate format
@@ -475,7 +569,11 @@ def construct_entries(
         exceptions = {}
         for converter in OPTIMADE_CONVERTERS[entry_config.entry_type]:
             try:
-                entry = converter(entry, properties=entry_config.property_definitions)  # type: ignore[call-arg]
+                entry = converter(
+                    entry,
+                    properties=entry_config.property_definitions,
+                    prefix=provider_prefix,
+                )  # type: ignore[call-arg]
                 if not isinstance(entry, dict):
                     entry = entry.entry
                 break
@@ -488,7 +586,7 @@ def construct_entries(
             )
 
         if not isinstance(entry, dict):
-            entry = entry.dict()
+            entry = entry.model_dump()
 
         if not entry["id"]:
             entry["id"] = unique_entry_id
@@ -506,12 +604,39 @@ def construct_entries(
             entry["attributes"]["immutable_id"] = file_path_entry_id
 
         entry["attributes"]["last_modified"] = timestamp
+    return optimade_entries
 
-    # Now try to parse the properties and assign them to OPTIMADE entries
-    _parse_and_assign_properties(
-        optimade_entries,
+
+def construct_entries(
+    archive_path: Path,
+    entry_config: EntryConfig,
+    provider_prefix: str,
+    limit: int | None = None,
+) -> dict[str, dict]:
+    if isinstance(entry_config.entry_paths, AiidaEntryPath):
+        optimade_entries = construct_entries_from_aiida(
+            archive_path, entry_config, provider_prefix
+        )
+
+    else:
+        optimade_entries = construct_entries_from_files(
+            archive_path, entry_config, provider_prefix, limit=limit
+        )
+
+    # Parse properties from `property_paths` and assign them to the OPTIMADE entries
+    property_matches_by_file: dict[str | None, list[Path]] = _get_matches(
+        archive_path, entry_config.property_paths
+    )
+    _check_missing(property_matches_by_file)
+
+    parsed_properties = _parse_properties_from_files(
         property_matches_by_file,
-        entry_config.entry_type,
+        entry_config.property_definitions,
+    )
+
+    _assign_and_validate_properties(
+        optimade_entries,
+        parsed_properties,
         entry_config.property_definitions,
         provider_prefix,
     )
@@ -525,6 +650,7 @@ def write_optimade_jsonl(
     property_definitions: dict[str, list[PropertyDefinition]],
     provider_prefix: str,
     jsonl_path: Path | None = None,
+    overwrite: bool = False,
 ) -> Path:
     """Write OPTIMADE entries to a JSONL file.
 
@@ -545,12 +671,12 @@ def write_optimade_jsonl(
     if not jsonl_path:
         jsonl_path = archive_path / "optimade.jsonl"
 
-    if jsonl_path.exists():
+    if jsonl_path.exists() and not overwrite:
         raise RuntimeError(f"Not overwriting existing file at {jsonl_path}")
 
-    with open(jsonl_path, "a") as jsonl:
+    with open(jsonl_path, "w") as jsonl:
         # write the optimade jsonl header
-        header = {"x-optimade": {"meta": {"api_version": "1.1.0"}}}
+        header = {"x-optimade": {"meta": {"api_version": OPTIMADE_API_VERSION}}}
         jsonl.write(json.dumps(header))
         jsonl.write("\n")
 
