@@ -3,7 +3,7 @@ import traceback
 from optimade_maker.config import UnsupportedConfigVersion
 
 archive_url = "https://staging-archive.materialscloud.org/"
-test_record_id = 1408
+test_record_id = 1419
 
 
 def test_archive_record_metadata():
@@ -12,7 +12,7 @@ def test_archive_record_metadata():
 
     try:
         record = ArchiveRecord(test_record_id, archive_url=archive_url)
-        assert len(record.files_w_checksums) == 4
+        assert len(record.files_w_checksums) == 5
         assert record.is_optimade_record() is True
     except UnsupportedConfigVersion:
         traceback.print_exc()
@@ -31,6 +31,6 @@ def test_archive_record_process():
         )
         record.process()
         files = os.listdir(record.default_path)
-        assert "structures.zip" in files
+        assert "structures.tar.gz" in files
     except UnsupportedConfigVersion:
         traceback.print_exc()
